from typing import Callable
from pyoxigraph import Store
from .db import Ingestable
Rule = Callable[[Store], Ingestable]

# it's possible to come up with an abstraction for a Rule
# but decided not to (here at least).

# an engine run (itself) could be taken as a 'rule'.


"""
just to use rdflib to canonicalize
"""
#use oxrdflib?
#also deal with strttl

class terms:
    def _():
        from rdflib.graph   import DATASET_DEFAULT_GRAPH_ID                   as DG
        from rdflib.term    import BNode     as BN, Literal as Lit, URIRef    as NN
        return locals()
    rdflib = _()
    def _():
        from pyoxigraph     import BlankNode as BN, Literal as Lit, NamedNode as NN
        from pyoxigraph     import DefaultGraph                               as DG; DG = DG()
        return locals()
    oxigraph = _()
    from types import SimpleNamespace as NS
    rdflib    = NS(**rdflib)
    oxigraph  = NS(**oxigraph)
    assert(frozenset(rdflib.__dict__.keys()) == frozenset(rdflib.__dict__.keys()))
    del NS

    class og2rl:
        def __call__(slf, n):
            rl = terms.rdflib
            og = terms.oxigraph
            if isinstance(n,
                        og.BN):
                return  rl.BN(n.value)
            elif isinstance(n,
                        og.Lit):
                return  rl.Lit(n.value,
                            datatype=None if n.language else slf(n.datatype),
                            #TypeError: A Literal can only have one of lang or datatype, per http://www.w3.org/TR/rdf-concepts/#section-Graph-Literal
                            lang=n.language)
            elif n ==   og.DG:
                return  rl.DG
            else:
                assert(isinstance(n, 
                        og.NN))
                return  rl.NN(n.value)
    og2rl = og2rl()

    class rl2og:
        def __call__(s, n):
            rl = terms.rdflib
            og = terms.oxigraph
            assert(isinstance(n, str)) # makes the following work for bn and nn
            if isinstance(n,
                        rl.BN):
                return  og.BN(n)
            elif isinstance(n,
                        rl.Lit):
                return  og.Lit(n,
                            datatype=s(n.datatype) if n.datatype else None,
                            language=n.language)
            if n ==     rl.DG:
                return  og.DG
            else:
                assert(isinstance(n,
                        rl.NN))
                return  og.NN(n)
    rl2og = rl2og()
terms = terms()


class types:
    from typing import Union
    rdflib =   Union[tuple(t for t in terms.rdflib  .__dict__.values() if t is not terms.rdflib  .DG)]
    oxigraph = Union[tuple(t for t in terms.oxigraph.__dict__.values() if t is not terms.oxigraph.DG)]

class rdflib:
    class oxigraph:
        from typing import Iterable
        from pyoxigraph import Triple, Quad
        def __call__(slf, d: Iterable[Iterable[types.oxigraph]]) -> Iterable[Iterable[types.rdflib]]:
            def _(q):
                _ = tuple(terms.rl2og(t) for t in q)
                if len(_) == 3: return slf.Triple(*_)
                else:
                    assert(len(_) == 4)
                    return slf.Quad(*_)
            _ = map(_, d)
            return _
            from pyoxigraph import serialize, RdfFormat
            _ = serialize(og, format=RdfFormat.N_QUADS)
            return _
            r = Dataset()
            r.parse(data=_, format='application/n-quads')
            return r
    oxigraph = oxigraph()
rdflib = rdflib()

class oxigraph:
    class rdflib:
        from typing import Iterable
        def __call__(slf, d: Iterable[Iterable[types.rdflib]]) -> Iterable[Iterable[types.oxigraph]]:
            # for putting it back in og
            _ = lambda q: tuple(terms.og2rl(t) for t in q)
            _ = map(_, d)
            return _
            _ = rl.serialize(format='application/n-quads')
            from pyoxigraph import parse, RdfFormat
            _ = parse(_, format=RdfFormat.N_QUADS)
            return _
    rdflib = rdflib()
oxigraph = oxigraph()



"""
canonicalization
"""


class _quads:
    from pyoxigraph import Quad, Triple, NamedNode
    from typing import Iterable


    def __call__(self, quads: Iterable[Quad]) -> Iterable[Quad]:
        """
        canonicalization of sets of quads
        """
        # the set of quads have to be broken into sets of triples
        # after canonicalization,
        # each set of triples has to be reassembled as quads
        from collections import defaultdict
        g2t = defaultdict(set)
        for q in quads: g2t[q.graph_name].add(q.triple)
        from .data import reification
        for g,ts in g2t.items():
            _ = ts
            _ = reification.standard(_)
            _ = triples(_)
            _ = reification.star(_)
            _ = map(lambda t: Quad(*t, graph_name=g), _)
            yield from _

    class _deanon:
        from pyoxigraph import Triple
        from typing import Iterable
        from pyoxigraph import Quad
        class defaults:
            uri = "urn:anon:hash:"
        def __call__(slf,
                quads: Iterable[Quad], *,
                uri = defaults.uri) -> Iterable[Quad]:
            """takes blank node value as an identifier for a uri"""
            _ = map(lambda q: slf.quad(q, uri), quads)
            return _
        @classmethod
        def quad(cls, q: Quad, uri):
            if isinstance(q.subject, cls.Triple):
                _ = cls.Triple(*(cls.f(n, uri) for n in q.subject))
                q = cls.Quad(_, q.predicate, q.object, q.graph_name)
            if isinstance(q.object, cls.Triple):
                _ = cls.Triple(*(cls.f(n, uri) for n in q.object))
                q = cls.Quad(q.subject, q.predicate, _, q.graph_name)
            return cls.Quad(*(cls.f(n, uri) for n in q))
        
        from pyoxigraph import BlankNode, NamedNode
        @classmethod
        def f(cls, n, uri: str):
            if isinstance(n, cls.BlankNode):
                return cls.NamedNode(uri+n.value)
            else:
                return n
    deanon = _deanon()
quads = _quads()


from pyoxigraph import BlankNode, Quad, Triple
from typing import Iterable
def hasanon(d: Iterable[Quad| Triple]):
    for q in d:
        if isinstance(q.subject, BlankNode):
            return True
        if isinstance(q.object, BlankNode):
            return True
    return False


def triples(ts):
    if not isinstance(ts, frozenset):
        ts = frozenset(ts)
    assert(isinstance(ts, frozenset))
    from pyoxigraph import Triple
    for t in ts:
        assert(not isinstance(t.subject,    Triple))
        assert(not isinstance(t.object,     Triple))
    # optimization
    if not hasanon(ts):
        return ts
    
    from . import conversions as c
    ts = c.oxigraph.rdflib(ts)
    from rdflib import Graph
    _ = Graph()
    for t in ts: ts = _.add(t)
    from rdflib.compare import to_canonical_graph
    ts = _; del _
    ts = to_canonical_graph(ts)
    ts = c.rdflib.oxigraph(ts)
    ts = frozenset(ts)
    return ts
def _ogtriples(triples):
    # algo seems to gets stuck (slow?)
    # wait for update TODO
    from pyoxigraph import Dataset, CanonicalizationAlgorithm, Quad
    def _(triples):
        d = Dataset(Quad(*t) for t in triples)
        d.canonicalize(CanonicalizationAlgorithm.UNSTABLE) # ?? unstable??
        for q in d: yield q.triple
    yield from _(triples)
